==> eav_field-2.0.0.tar.gz: eav_field/src/Entity/EavAttribute.php <==
<?php

namespace Drupal\eav_field\Entity;

use Drupal\Core\Annotation\Translation;
use Drupal\Core\Entity\Annotation\ContentEntityType;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManager;
use Drupal\Core\Field\FormatterPluginManager;
use Drupal\Core\Field\WidgetPluginManager;
use Drupal\eav_field\EavAttributeStorage;

/**
 * @ContentEntityType(
 *   id = "eav_attribute",
 *   label = @Translation("EAV attribute"),
 *   base_table = "eav_attribute",
 *   data_table = "eav_attribute_field_data",
 *   entity_keys = {
 *     "id" = "aid",
 *     "uuid" = "uuid",
 *     "label" = "label",
 *     "langcode" = "langcode",
 *     "weight" = "weight",
 *   },
 *   translatable = TRUE,
 *   fieldable = TRUE,
 *   field_ui_base_route = "entity.eav_attribute.collection",
 *   handlers = {
 *     "storage" = "\Drupal\eav_field\EavAttributeStorage",
 *     "form" = {
 *       "add" = "\Drupal\eav_field\Form\EavAttributeForm",
 *       "edit" = "\Drupal\eav_field\Form\EavAttributeForm",
 *       "delete" = "\Drupal\Core\Entity\ContentEntityDeleteForm",
 *     },
 *     "list_builder" = "\Drupal\eav_field\EavAttributeListBuilder",
 *   },
 *   links = {
 *     "add-form" = "/admin/structure/eav/attributes/add",
 *     "edit-form" = "/admin/structure/eav/attributes/{eav_attribute}/edit",
 *     "delete-form" = "/admin/structure/eav/attributes/{eav_attribute}/delete",
 *     "collection" = "/admin/structure/eav/attributes",
 *   },
 *   admin_permission = "administer eav attributes",
 * )
 */
class EavAttribute extends ContentEntityBase {

  /**
   * {@inheritDoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type); /** @var BaseFieldDefinition[] $fields */
    $i = 0;

    $fields['aid']
      ->setLabel(t('EAV attribute ID'))
      ->setDescription(t('EAV attribute ID'));

    $fields['label'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Label'))
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['admin_label'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Administrative label'))
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['machine_name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Machine name'))
      ->setRequired(TRUE)
      ->addConstraint('UniqueField')
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $allowed_field_types = array_keys(EavValue::getFieldTypes());
    $fields['value_type'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('Value field type'))
      ->setRequired(TRUE)
      ->setDefaultValue('string')
      ->setSettings([
        'allowed_values' => array_combine($allowed_field_types, $allowed_field_types),
      ])
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['value_storage_config'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Value storage config'));

    $fields['value_config'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Value field config'));

    $fields['value_widget_type'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('Widget'))
      ->setRequired(TRUE)
      ->setDefaultValue('string_textfield')
      ->setSetting('allowed_values_function', [static::class, 'getValueWidgetTypeAllowedValues'])
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['value_widget_settings'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Widget settings'));

    $fields['value_formatter_type'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('Formatter'))
      ->setSetting('allowed_values_function', [static::class, 'getValueFormatterTypeAllowedValues'])
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['value_formatter_settings'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Formatter settings'));

    $fields['description'] = BaseFieldDefinition::create('text_long')
      ->setLabel(t('Description'))
      ->setDisplayOptions('form', [
        'type' => 'text_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['primary'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Is primary attribute'))
      ->setDefaultValue(FALSE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['weight'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Weight'))
      ->setDisplayOptions('form', [
        'type' => 'number',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    return $fields;
  }

  /**
   * {@inheritDoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    if ($this->getWeight() === NULL) {
      $this->set('weight', self::getMaxWeight() + 1);
    }
  }

  /**
   * Load attribute by id.
   *
   * @return EavAttribute|null
   */
  public static function load($id) {
    return parent::load($id);
  }

  /**
   * Load attribute by machine name.
   *
   * @return EavAttribute|null
   */
  public static function loadByMachineName($machine_name) {
    $attribute_storage = \Drupal::entityTypeManager()->getStorage('eav_attribute'); /** @var EavAttributeStorage $attribute_storage */
    $attributes = $attribute_storage->loadByProperties(['machine_name' => $machine_name]);
    return $attributes ? current($attributes) : NULL;
  }

  /**
   * Load unchanged attribute.
   *
   * @return EavAttribute
   */
  public function loadUnchanged() {
    /** @var EavAttribute $unchanged_attribute */
    $unchanged_attribute = \Drupal::entityTypeManager()->getStorage('eav_attribute')->loadUnchanged($this->id());
    return $unchanged_attribute;
  }

  /**
   * Return administrative label.
   *
   * @return string
   */
  public function getAdministrativeLabel() {
    return $this->get('admin_label')->value ?: $this->label();
  }

  /**
   * Return attribute machine name.
   *
   * @return string
   */
  public function getMachineName() {
    return $this->get('machine_name')->value;
  }

  /**
   * Return attribute weight.
   *
   * @return integer
   */
  public function getWeight() {
    return $this->get('weight')->value;
  }

  /**
   * Return "value field" type.
   *
   * @return string
   */
  public function getValueType() {
    return $this->get('value_type')->value;
  }

  /**
   * Return "value field" name.
   *
   * @return string
   */
  public function getValueFieldName() {
    return $this->getValueType() . '_value';
  }

  /**
   * Return "value field" storage config array.
   *
   * @return array
   */
  public function getValueStorageConfigArray($with_defaults = TRUE) {
    $value_storage_config_array = $this->get('value_storage_config')->getValue()[0] ?? [];

    if ($with_defaults) {
      $default_values = [
        'field_name' => $this->getValueFieldName(),
        'entity_type' => 'eav_value',
        'type' => $this->getValueType(),
      ];

      $field_type_manager = \Drupal::service('plugin.manager.field.field_type'); /** @var FieldTypePluginManager $field_type_manager */
      $default_settings = $field_type_manager->getDefaultStorageSettings($this->getValueType());

      $value_storage_config_array = $default_values + $value_storage_config_array + ['settings' => $default_settings];
    }

    return $value_storage_config_array;
  }

  /**
   * Return "value field" config array.
   *
   * @return array
   */
  public function getValueConfigArray($with_defaults = TRUE) {
    $value_config_array = $this->get('value_config')->getValue()[0] ?? [];

    if ($with_defaults) {
      $default_values = [
        'field_name' => $this->getValueFieldName(),
        'entity_type' => 'eav_value',
        'bundle' => 'eav_value',
      ];

      $field_type_manager = \Drupal::service('plugin.manager.field.field_type'); /** @var FieldTypePluginManager $field_type_manager */
      $default_settings = $field_type_manager->getDefaultFieldSettings($this->getValueType());

      $value_config_array = $default_values + $value_config_array + ['settings' => $default_settings];
    }

    return $value_config_array;
  }

  /**
   * Return value field default values.
   */
  public function getValueFieldDefaultValues(): ?array {
    return $this->getValueConfigArray()['default_value'] ?? NULL;
  }

  /**
   * Return "value field" widget type.
   *
   * @return string
   */
  public function getValueWidgetType() {
    return $this->get('value_widget_type')->value;
  }

  /**
   * Return "value field" widget settings.
   *
   * @return array
   */
  public function getValueWidgetSettings() {
    return $this->get('value_widget_settings')->getValue()[0] ?? [];
  }

  /**
   * Return "value field" formatter type.
   *
   * @return string
   */
  public function getValueFormatterType() {
    return $this->get('value_formatter_type')->value;
  }

  /**
   * Return "value field" formatter settings.
   *
   * @return array
   */
  public function getValueFormatterSettings() {
    return $this->get('value_formatter_settings')->getValue()[0] ?? [];
  }

  /**
   * Return TRUE if attribute is primary.
   *
   * @return boolean
   */
  public function isPrimary() {
    return $this->get('primary')->value;
  }

  /**
   * Return value field definition.
   *
   * @return BaseFieldDefinition
   */
  public function getValueFieldDefinition() {
    $entity_field_manager = \Drupal::service('entity_field.manager'); /** @var EntityFieldManagerInterface $entity_field_manager */
    return $entity_field_manager->getFieldDefinitions('eav_value', 'eav_value')[$this->getValueFieldName()];
  }

  /**
   * Configure value field definition and return him definition.
   *
   * @return BaseFieldDefinition
   */
  public function configureValueFieldDefinition() {
    $field_definition = $this->getValueFieldDefinition();
    $field_storage_definition = $field_definition->getFieldStorageDefinition();
    $field_config_array = $this->getValueConfigArray();
    $field_storage_config_array = $this->getValueStorageConfigArray();

    // Configure field storage
    if (!empty($field_storage_config_array['settings'])) {
      $field_storage_definition->setSettings($field_storage_config_array['settings']);
    }
    $field_storage_definition->setCardinality($field_storage_config_array['cardinality'] ?? 1);

    // Configure field
    $field_definition->setLabel($this->label());
    if ($field_storage_definition->getCardinality() != 1) {
      $field_definition->setLabel(t('Values'));
    }
    if (!empty($field_config_array['settings'])) {
      $field_definition->setSettings($field_config_array['settings']);
    }
    if (isset($field_config_array['default_value'])) {
      $field_definition->setDefaultValue($field_config_array['default_value']);
    }
    if (isset($field_config_array['description'])) {
      $field_definition->setDescription($field_config_array['description']);
    }
    if (isset($field_config_array['required'])) {
      $field_definition->setRequired($field_config_array['required']);
    }

    $options_allowed_values_static_cache = &drupal_static('options_allowed_values', []);
    $valie_field_name = $this->getValueFieldName();
    unset(
      $options_allowed_values_static_cache['eav_value:' . $valie_field_name],
      $options_allowed_values_static_cache['eav_value:' . $valie_field_name . ':entity']
    );

    return $field_definition;
  }

  /**
   * Return categories ids.
   *
   * @return integer[]
   */
  public function getCategoriesIds() {
    $categories_ids = [];

    foreach ($this->get('category')->getValue() as $value) {
      $categories_ids[] = $value['target_id'];
    }

    return $categories_ids;
  }

  /**
   * Return EAV attributes max weight.
   *
   * @return integer
   */
  public static function getMaxWeight() {
    $query = \Drupal::database()->select('eav_attribute_field_data', 'a');
    $query->addExpression('MAX(a.weight)');
    return (int)$query->execute()->fetchField();
  }

  /**
   * Return "value widget type" field allowed values.
   */
  public static function getValueWidgetTypeAllowedValues(FieldStorageDefinitionInterface $field_storage_definition, EavAttribute $attribute = NULL, &$cacheable = TRUE) {
    $cacheable = FALSE;

    $widget_manager = \Drupal::service('plugin.manager.field.widget'); /** @var WidgetPluginManager $widget_manager */
    $field_type = $attribute ? $attribute->getValueType() : NULL;
    return $widget_manager->getOptions($field_type);
  }

  /**
   * Return "value formatter type" field allowed values.
   */
  public static function getValueFormatterTypeAllowedValues(FieldStorageDefinitionInterface $field_storage_definition, EavAttribute $attribute = NULL, &$cacheable = TRUE) {
    $cacheable = FALSE;

    $formatter_manager = \Drupal::service('plugin.manager.field.formatter'); /** @var FormatterPluginManager $formatter_manager */
    $field_type = $attribute ? $attribute->getValueType() : NULL;
    return $formatter_manager->getOptions($field_type);
  }

}

==> eav_field-2.0.1.tar.gz: eav_field/src/Entity/EavAttribute.php <==
<?php

namespace Drupal\eav_field\Entity;

use Drupal\Core\Annotation\Translation;
use Drupal\Core\Entity\Annotation\ContentEntityType;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManager;
use Drupal\Core\Field\FormatterPluginManager;
use Drupal\Core\Field\WidgetPluginManager;
use Drupal\eav_field\EavAttributeStorage;

/**
 * @ContentEntityType(
 *   id = "eav_attribute",
 *   label = @Translation("EAV attribute"),
 *   base_table = "eav_attribute",
 *   data_table = "eav_attribute_field_data",
 *   entity_keys = {
 *     "id" = "aid",
 *     "uuid" = "uuid",
 *     "label" = "label",
 *     "langcode" = "langcode",
 *     "weight" = "weight",
 *   },
 *   translatable = TRUE,
 *   fieldable = TRUE,
 *   field_ui_base_route = "entity.eav_attribute.collection",
 *   handlers = {
 *     "storage" = "\Drupal\eav_field\EavAttributeStorage",
 *     "form" = {
 *       "add" = "\Drupal\eav_field\Form\EavAttributeForm",
 *       "edit" = "\Drupal\eav_field\Form\EavAttributeForm",
 *       "delete" = "\Drupal\Core\Entity\ContentEntityDeleteForm",
 *     },
 *     "list_builder" = "\Drupal\eav_field\EavAttributeListBuilder",
 *   },
 *   links = {
 *     "add-form" = "/admin/structure/eav/attributes/add",
 *     "edit-form" = "/admin/structure/eav/attributes/{eav_attribute}/edit",
 *     "delete-form" = "/admin/structure/eav/attributes/{eav_attribute}/delete",
 *     "collection" = "/admin/structure/eav/attributes",
 *   },
 *   admin_permission = "administer eav attributes",
 * )
 */
class EavAttribute extends ContentEntityBase {

  /**
   * {@inheritDoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type); /** @var BaseFieldDefinition[] $fields */
    $i = 0;

    $fields['aid']
      ->setLabel(t('EAV attribute ID'))
      ->setDescription(t('EAV attribute ID'));

    $fields['label'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Label'))
      ->setRequired(TRUE)
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['admin_label'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Administrative label'))
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['machine_name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Machine name'))
      ->setRequired(TRUE)
      ->addConstraint('UniqueField')
      ->setDisplayOptions('form', [
        'type' => 'string_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $allowed_field_types = array_keys(EavValue::getFieldTypes());
    $fields['value_type'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('Value field type'))
      ->setRequired(TRUE)
      ->setDefaultValue('string')
      ->setSettings([
        'allowed_values' => array_combine($allowed_field_types, $allowed_field_types),
      ])
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['value_storage_config'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Value storage config'));

    $fields['value_config'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Value field config'));

    $fields['value_widget_type'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('Widget'))
      ->setRequired(TRUE)
      ->setDefaultValue('string_textfield')
      ->setSetting('allowed_values_function', [static::class, 'getValueWidgetTypeAllowedValues'])
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['value_widget_settings'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Widget settings'));

    $fields['value_formatter_type'] = BaseFieldDefinition::create('list_string')
      ->setLabel(t('Formatter'))
      ->setSetting('allowed_values_function', [static::class, 'getValueFormatterTypeAllowedValues'])
      ->setDisplayOptions('form', [
        'type' => 'options_select',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['value_formatter_settings'] = BaseFieldDefinition::create('map')
      ->setLabel(t('Formatter settings'));

    $fields['description'] = BaseFieldDefinition::create('text_long')
      ->setLabel(t('Description'))
      ->setDisplayOptions('form', [
        'type' => 'text_textfield',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['primary'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Is primary attribute'))
      ->setDefaultValue(FALSE)
      ->setDisplayOptions('form', [
        'type' => 'boolean_checkbox',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    $fields['weight'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Weight'))
      ->setDisplayOptions('form', [
        'type' => 'number',
        'weight' => $i++,
      ])
      ->setDisplayConfigurable('form', TRUE);

    return $fields;
  }

  /**
   * {@inheritDoc}
   */
  public function preSave(EntityStorageInterface $storage) {
    parent::preSave($storage);

    if ($this->getWeight() === NULL) {
      $this->set('weight', self::getMaxWeight() + 1);
    }
  }

  /**
   * Load attribute by id.
   *
   * @return EavAttribute|null
   */
  public static function load($id) {
    /** @noinspection PhpIncompatibleReturnTypeInspection */
    return parent::load($id);
  }

  /**
   * Load attribute by machine name.
   */
  public static function loadByMachineName($machine_name): ?EavAttribute {
    $attribute_storage = \Drupal::entityTypeManager()->getStorage('eav_attribute'); /** @var EavAttributeStorage $attribute_storage */
    $attributes = $attribute_storage->loadByProperties(['machine_name' => $machine_name]);
    return $attributes ? current($attributes) : NULL;
  }

  /**
   * Load unchanged attribute.
   */
  public function loadUnchanged(): ?EavAttribute {
    /** @noinspection PhpIncompatibleReturnTypeInspection */
    return $this->entityTypeManager()->getStorage('eav_attribute')->loadUnchanged($this->id());
  }

  /**
   * Return administrative label.
   */
  public function getAdministrativeLabel(): string {
    return $this->get('admin_label')->value ?: $this->label();
  }

  /**
   * Return attribute machine name.
   */
  public function getMachineName(): string {
    return $this->get('machine_name')->value;
  }

  /**
   * Return attribute weight.
   */
  public function getWeight(): ?int {
    return $this->get('weight')->value;
  }

  /**
   * Return "value field" type.
   */
  public function getValueType(): string {
    return $this->get('value_type')->value;
  }

  /**
   * Return "value field" name.
   */
  public function getValueFieldName(): string {
    return $this->getValueType() . '_value';
  }

  /**
   * Return "value field" storage config array.
   */
  public function getValueStorageConfigArray($with_defaults = TRUE): array {
    $value_storage_config_array = $this->get('value_storage_config')->getValue()[0] ?? [];

    if ($with_defaults) {
      $default_values = [
        'field_name' => $this->getValueFieldName(),
        'entity_type' => 'eav_value',
        'type' => $this->getValueType(),
      ];

      $field_type_manager = \Drupal::service('plugin.manager.field.field_type'); /** @var FieldTypePluginManager $field_type_manager */
      $default_settings = $field_type_manager->getDefaultStorageSettings($this->getValueType());

      $value_storage_config_array = $default_values + $value_storage_config_array + ['settings' => $default_settings];
    }

    return $value_storage_config_array;
  }

  /**
   * Return "value field" config array.
   */
  public function getValueConfigArray($with_defaults = TRUE): array {
    $value_config_array = $this->get('value_config')->getValue()[0] ?? [];

    if ($with_defaults) {
      $default_values = [
        'field_name' => $this->getValueFieldName(),
        'entity_type' => 'eav_value',
        'bundle' => 'eav_value',
      ];

      $field_type_manager = \Drupal::service('plugin.manager.field.field_type'); /** @var FieldTypePluginManager $field_type_manager */
      $default_settings = $field_type_manager->getDefaultFieldSettings($this->getValueType());

      $value_config_array = $default_values + $value_config_array + ['settings' => $default_settings];
    }

    return $value_config_array;
  }

  /**
   * Return value field default values.
   */
  public function getValueFieldDefaultValues(): ?array {
    return $this->getValueConfigArray()['default_value'] ?? NULL;
  }

  /**
   * Return "value field" widget type.
   */
  public function getValueWidgetType(): string {
    return $this->get('value_widget_type')->value;
  }

  /**
   * Return "value field" widget settings.
   */
  public function getValueWidgetSettings(): array {
    return $this->get('value_widget_settings')->getValue()[0] ?? [];
  }

  /**
   * Return "value field" formatter type.
   */
  public function getValueFormatterType(): string {
    return $this->get('value_formatter_type')->value;
  }

  /**
   * Return "value field" formatter settings.
   */
  public function getValueFormatterSettings(): array {
    return $this->get('value_formatter_settings')->getValue()[0] ?? [];
  }

  /**
   * Return TRUE if attribute is primary.
   */
  public function isPrimary(): bool {
    return (bool)$this->get('primary')->value;
  }

  /**
   * Return value field definition.
   */
  public function getValueFieldDefinition(): FieldDefinitionInterface {
    $entity_field_manager = \Drupal::service('entity_field.manager'); /** @var EntityFieldManagerInterface $entity_field_manager */
    return $entity_field_manager->getFieldDefinitions('eav_value', 'eav_value')[$this->getValueFieldName()];
  }

  /**
   * Configure value field definition and return him definition.
   *
   * @return BaseFieldDefinition
   */
  public function configureValueFieldDefinition() {
    $field_definition = $this->getValueFieldDefinition();
    $field_storage_definition = $field_definition->getFieldStorageDefinition();
    $field_config_array = $this->getValueConfigArray();
    $field_storage_config_array = $this->getValueStorageConfigArray();

    // Configure field storage
    if (!empty($field_storage_config_array['settings'])) {
      $field_storage_definition->setSettings($field_storage_config_array['settings']);
    }
    $field_storage_definition->setCardinality($field_storage_config_array['cardinality'] ?? 1);

    // Configure field
    $field_definition->setLabel($this->label());
    if ($field_storage_definition->getCardinality() != 1) {
      $field_definition->setLabel(t('Values'));
    }
    if (!empty($field_config_array['settings'])) {
      $field_definition->setSettings($field_config_array['settings']);
    }
    if (isset($field_config_array['default_value'])) {
      $field_definition->setDefaultValue($field_config_array['default_value']);
    }
    if (isset($field_config_array['description'])) {
      $field_definition->setDescription($field_config_array['description']);
    }
    if (isset($field_config_array['required'])) {
      $field_definition->setRequired($field_config_array['required']);
    }

    $options_allowed_values_static_cache = &drupal_static('options_allowed_values', []);
    $valie_field_name = $this->getValueFieldName();
    unset(
      $options_allowed_values_static_cache['eav_value:' . $valie_field_name],
      $options_allowed_values_static_cache['eav_value:' . $valie_field_name . ':entity']
    );

    return $field_definition;
  }

  /**
   * Return categories ids.
   *
   * @return integer[]
   */
  public function getCategoriesIds(): array {
    $categories_ids = [];

    foreach ($this->get('category')->getValue() as $value) {
      $categories_ids[] = $value['target_id'];
    }

    return $categories_ids;
  }

  /**
   * Return EAV attributes max weight.
   */
  public static function getMaxWeight(): int {
    $query = \Drupal::database()->select('eav_attribute_field_data', 'a');
    $query->addExpression('MAX(a.weight)');
    return (int)$query->execute()->fetchField();
  }

  /**
   * Return "value widget type" field allowed values.
   */
  public static function getValueWidgetTypeAllowedValues(FieldStorageDefinitionInterface $field_storage_definition, EavAttribute $attribute = NULL, &$cacheable = TRUE): array {
    $cacheable = FALSE;

    $widget_manager = \Drupal::service('plugin.manager.field.widget'); /** @var WidgetPluginManager $widget_manager */
    $field_type = $attribute ? $attribute->getValueType() : NULL;
    return $widget_manager->getOptions($field_type);
  }

  /**
   * Return "value formatter type" field allowed values.
   */
  public static function getValueFormatterTypeAllowedValues(FieldStorageDefinitionInterface $field_storage_definition, EavAttribute $attribute = NULL, &$cacheable = TRUE): array {
    $cacheable = FALSE;

    $formatter_manager = \Drupal::service('plugin.manager.field.formatter'); /** @var FormatterPluginManager $formatter_manager */
    $field_type = $attribute ? $attribute->getValueType() : NULL;
    return $formatter_manager->getOptions($field_type);
  }

}
